  /* example_toggle that hides >>>, ..., etc */
.example_toggle {
    cursor: pointer;
    position: absolute;
    top: 0px;
    right: 0px;
    border: 1px solid rgb(221, 221, 221);
    color: rgb(221, 221, 221);
    font-family: monospace;
    padding-left: 0.2rem;
    padding-right: 0.2rem;
  }

  div.highlight:hover span.example_toggle::after {
    background: #3F556B;
    border-radius: 0.25rem;
    color: white;
    content: attr(title);
    padding: 0.25rem;
    position: absolute;
    z-index: 98;
    width: 100px;
    font-size: 0.7rem;
    top: 0;
    right: 0;
  }
