# -*- coding: utf-8 -*-
"""Lists of datasets provided in this directory."""

classification_data = [
    "ACSF1",
    "ArrowHead",
    "BasicMotions",
    "ItalyPowerDemand",
    "OSULeaf",
    "PLAID",
]
