# -*- coding: utf-8 -*-
"""Convolution based time series classifiers."""

__all__ = ["RocketClassifier", "Arsenal"]

from aeon.classification.convolution_based._arsenal import Arsenal
from aeon.classification.convolution_based._rocket_classifier import RocketClassifier
