# -*- coding: utf-8 -*-
"""Series as features - interval based."""
__all__ = ["BaseTimeSeriesForest"]

from aeon.series_as_features.base.estimators.interval_based._tsf import (
    BaseTimeSeriesForest,
)
