# -*- coding: utf-8 -*-
"""Signature based panel transformations."""
__all__ = [
    "SignatureTransformer",
]

from aeon.transformations.panel.signature_based._signature_method import (
    SignatureTransformer,
)
