# -*- coding: utf-8 -*-
"""Module for general partitioning algorithms."""
from aeon.clustering.partitioning._lloyds import TimeSeriesLloyds

__all__ = ["TimeSeriesLloyds"]
__author__ = ["chrisholder", "TonyBagnall"]
