# -*- coding: utf-8 -*-
"""Implement composite time series regression estimators."""

__all__ = [
    "RegressorPipeline",
    "SklearnRegressorPipeline",
]

from aeon.regression.compose._pipeline import (
    RegressorPipeline,
    SklearnRegressorPipeline,
)
